#!/bin/bash
PRISM_HOME=`pwd` #this is the root directory of the PRISM_Project, which should also be the directory containing this script

if [ $# -lt 4 ]
then
    echo "### missing parameter(s) ###"
    echo "Usage $0 <word> <input_file> <exponent for t> <number of rounds> [<secret_key>]"
    exit
fi

bash build_PRISM_Jar.sh #Build the Jar file necessary to run the job

word=$1
input_file_path=$2

input_file=`echo $input_file_path | awk -F '/' '{print $NF }'` #we take only the filename which is the last element of the path
day=`echo $input_file | cut -d "-" -f1,2,3` #e.g echo 16-01-2011-hello.txt | cut -d "-" -f1,2,3 --> 16-01-2011
echo "day recovered from input file name: "$day

t_exponent=$3
num_rounds=$4

test_output='output/'$input_file'_q'$num_rounds'_t'$t_exponent

current_counter=0
flag_exist=0
while [ $flag_exist -eq 0 ] 
do
    #echo "test output:"$test_output'_cnt'$current_counter
    echo "Testing file existence on HDFS..."
    hadoop fs -test -e $test_output'_cnt'$current_counter #while the target destination folder exist,NB if the folder exist the test command returns 0
    flag_exist=$? #return value of previous command
    
    if [ $flag_exist -eq 0 ]
    then
      let 'current_counter=current_counter+1'
    fi

done

echo "counter to append to output directory : "$current_counter

output_file='output/'$input_file'_q'$num_rounds'_t'$t_exponent'_cnt'$current_counter


if [ $# -ge 5 ] #if there is a 5th parameter it should be the secret_key key
then
  secret_key=$5
else
  secret_key="DEFAULT"
fi

echo "Starting MapReduce..."
export HADOOP_CLASSPATH=$HADOOP_CLASSPATH:./libs/*:./libs/gnu-crypto-2.0.1-bin/lib/*

hadoop jar ./PRISM_Jars/PRISM.jar Search -libjars ./libs/bcprov-jdk16-145.jar,./libs/gnu-crypto-2.0.1-bin/lib/gnu-crypto.jar,./libs/gnu-crypto-2.0.1-bin/lib/javax-crypto.jar,./libs/gnu-crypto-2.0.1-bin/lib/javax-security.jar $word $input_file_path $output_file $t_exponent $num_rounds $secret_key

#cleanup local files created
rm local_alphas_values.dat 2>/dev/null
rm local_E_Pinf_value.dat 2>/dev/null

echo "Search completed, now execute 'bash decode_PRISMOutput.sh' on the output_directory located on HDFS $output_file to see the results"  



#examples

#./search.sh foo /user/edy/Enc_Input/20-01-2011-input.txt /user/edy/out/search.out

#./search.sh Hello /user/edy/Enc_Input/17-01-2011-hello.txt /user/edy/out/search.out
