import java.math.BigInteger;

import gnu.crypto.cipher.IBlockCipher;


import java.io.*;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Date;

import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.WritableUtils;



/** The purpose of this class is to decrypt the output of the search algorithm*/
public class DecryptSearchOutput {
	
	private static byte keyValueSeparator = "|".getBytes()[0]; // "|" is exactly 1 byte
	
	public static void main(String[] args) throws Exception {
		   
	   	
	   	if(args.length < 7){
	   		 System.out.println("Error Missing parameters");
			 System.out.println("Usage: DecryptSearchOutput <word> <input_ciphered_file>  <ouput_plain_text_file> <exponent for t> <number of rounds> <day> <secret_key>");
			 System.exit(1);
		}
	   	
	   	String word=args[0];
	   	word=word.trim();
	   	
	   	int total_rows=0;
	   	int num_rounds=0;
		int t_exp=0;
		int X_Y_LENGTH = 0;
		
		String exp_str=args[3]; // value of t's exponent, remember t= 2^e
		String q_str = args[4]; //#rounds q
		String day_str=args[5];
		String secret_key=args[6];


		//SAMPLE
		//BigInteger p = new BigInteger("83");
		//BigInteger N = new BigInteger("2");
		//BigInteger b = new BigInteger("33");
		//BigInteger b_inverse = b.modInverse(p);

		BigInteger p = new BigInteger("1564820277778863089963405534144523812327502518356860558287169907464574863492764499627171260275816613599242003651320891327");
		BigInteger N = new BigInteger("20");
		BigInteger b = new BigInteger("1298551010326079989062112961798042817861449001333409542125059506044661625781468711882345844134701566170769791452428186529");
		BigInteger b_inverse = b.modInverse(p);

		
		try{
			num_rounds=Integer.parseInt(q_str);
			t_exp=Integer.parseInt(exp_str);
		}catch(NumberFormatException e){
			System.err.println("Error q and t must be integer values! Aborting!");
			System.exit(1);
		}
		
		total_rows=(int)Math.pow(2, t_exp);
		X_Y_LENGTH = 2*(int)PRISM_Utils.log2(total_rows); //== 2*log2(t) length of the concatenation Xi||Yi
		
	   	RIJNDAEL_Cipher RIJ = null;
		
		if(secret_key.equals("DEFAULT")){
			RIJ= new RIJNDAEL_Cipher(day_str); //This will use the default key string for our RIJNDAEL Cipher
		}else{
			RIJ= new RIJNDAEL_Cipher(secret_key,day_str);
		}
	   	
	   	IBlockCipher rij_cipher = RIJ.getCipher();
		int blocksize = rij_cipher.currentBlockSize();
		
		ArrayList<Integer> rounds_to_check =  new ArrayList<Integer>();
		String row_to_check= "";
		
		byte[] plain_bytes= PRISM_Utils.concatenate(word.getBytes(), 1);
		byte[] padded_bytes = RIJNDAEL_Cipher.padBytes(plain_bytes, blocksize);
		byte[] encrypted = new byte[padded_bytes.length];
		rij_cipher.encryptBlock(padded_bytes, 0, encrypted, 0);
		int [] x_y = Search.get_candidate_x_y(encrypted,X_Y_LENGTH);
		int x = x_y[0];
		
		row_to_check = x<10?"0"+x:x+""; // if the row number is smaller than 10, I put a '0' in front, this is because 
										//rows are sorted alphabetically at the output of the reducer, therefore row 2 would come after row 19 without this trick
	
		
		// I compute the rounds for which the user expects a 1
		for(int i =0;i<num_rounds;i++){
		   	
		   	
		    if(PRISM_Utils.getBit(encrypted,X_Y_LENGTH+i)==1){ 
				rounds_to_check.add(i); // this are the rounds for which a user expects a 1 in other words h( E(w,1),round_num) = 1 (the first bit)
			} 
	   	}
	   	
	   	DataInputStream in = new DataInputStream(new FileInputStream(args[1]));
	   	
		File file_output = new File(args[2]);
		//BufferedWriter out = new BufferedWriter(new FileWriter(file_output));

					
		boolean done_header = false;
		int count =0;
		String row_output="";
		boolean contradiction = false;

	
		while(count < total_rows)
		{
			row_output="";
			Text key = new Text();
			key.readFields(in);
			String str = key.toString();
			//System.out.println();
			if(!done_header){
			        System.out.println("Overview: "+str);
				System.out.print("====PRISM SEARCH====\n");
				System.out.print("Searched expression : "+word+"\n");
				String file = str.split(",")[0];
				System.out.print("In file:"+file+"\n");
				System.out.print("number of rounds :"+num_rounds+"\n");
				System.out.print("number of rows :"+total_rows+"\n");
				System.out.print("-------RESULTS--------\n");
				System.out.print("row \t (round,PIR value) \n");
				done_header = true;
			}
			String row_num = str.split(",")[1];
			
			byte KVseparator = in.readByte();
		      			
			//System.out.println("separator:"+(char)KVseparator);


			BytesArrayWritable rec = new BytesArrayWritable();
			rec.readFields(in);
			BigInteger[] sum = Search.decodeBytesArray(rec);
	
			
			if(row_num.equals(row_to_check)){
				System.out.print(row_to_check+"\t");
				for(int rnd:rounds_to_check){
				
					sum[rnd] = (sum[rnd].multiply(b_inverse)).mod(p);
					int result = (sum[rnd].mod(N)).intValue();
				        //System.out.println("Returned "+result+" in round "+rnd);
					if(result==1){
						row_output+="("+rnd+",1)";
					}else if(result==0){
						row_output+="("+rnd+",0)";
						contradiction = true; 
					  
					}else{ // this is the case where I get E(1)+E(1)+E(0)+...+E(1) = Enc(n) witn n > 1
						row_output+="("+rnd+",n,), n > 1)";
					}
				}
				
				System.out.print(row_output+"\n");
				
				}
			count++;
		}
		
		if(contradiction){ 
			System.out.print("CONTRADICTION !! I was expecting a 1 in the above rounds and I got a 0 in one of the rounds\n");
		}else{
			System.out.print("NO CONTRADICTION found after "+num_rounds+ " rounds, the expression may probably be present in the file\n");
			}
	   	in.close();
	   	//out.close();
	}

}
