import java.io.IOException;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FSDataInputStream;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.BytesWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.compress.CompressionCodec;
import org.apache.hadoop.io.compress.CompressionCodecFactory;
import org.apache.hadoop.mapreduce.InputSplit;
import org.apache.hadoop.mapreduce.RecordReader;
import org.apache.hadoop.mapreduce.TaskAttemptContext;
import org.apache.hadoop.mapreduce.lib.input.FileSplit;

//Erik
//import org.apache.commons.logging.LogFactory;
//import org.apache.commons.logging.Log;

/**
 * Treats keys as offset of encrypted words in file and value as encrypted word.
 * I inspired myself from the LineRecordReader that is used to read a Text :
 * I read blocks of WORD_LENGTH bytes then I take their offset in the file as the key and the block itself as the value 
 */


// I should have a look at : SequenceFileAsBinaryInputFormat.java 
// check : http://hadoop.sourcearchive.com/documentation/0.20.2plus-pdfsg1-2/SequenceFileAsBinaryInputFormat_8java-source.html

//ALSO: http://www.java2s.com/Code/Java/File-Input-Output/CopyBytesbetweenFileInputStreamandFileOutputStream.htm
public class BytesRecordReader extends RecordReader<LongWritable, BytesWritable>{
		
	  /**
	   * KEYOUT=LongWritable : the position of a bloc in the ciphered text
	   * VALUE OUT = BytesWritable : the ciphered bloc
	   */
          //Erik
	  //private static final Log LOG = LogFactory.getLog(BytesRecordReader.class);
	  
	  private CompressionCodecFactory compressionCodecs = null;
	  private long start;
	  
	  //Erik private long pos=0;
	  private long end;
	  int bytes_read=0;
	  long marker=0;//I will use this to compute the key
	  
	  //private LineReader in;
	  FSDataInputStream fileIn=null;
	  private int maxLineLength;
	  
	  private LongWritable key = null;//This is the position of each block read in the inputSplit
	  private BytesWritable value = null;//Used to store a ciphered bloc of bytes
	  
	  byte[] buf = new byte[PRISM_Settings.RIJNDAEL_BLOCK_SIZE];

	  public void initialize(InputSplit genericSplit,
	                         TaskAttemptContext context) throws IOException {
		  
	    FileSplit split = (FileSplit) genericSplit;
	    Configuration job = context.getConfiguration();
	    this.maxLineLength = job.getInt("mapred.linerecordreader.maxlength",
	                                    Integer.MAX_VALUE);
	    
	    // I SHOULD REMOVE UNECESSARY LOGGING WHEN DONE WITH DEBUGGING
	   
	    start = split.getStart();
	    
	    end = start + split.getLength();

	    //Erik
	    System.out.println("Split starts at:"+start + ", split ends at: "+end);
	    
	    final Path file = split.getPath();
	    compressionCodecs = new CompressionCodecFactory(job);
	    final CompressionCodec codec = compressionCodecs.getCodec(file);

	    // open the file and seek to the start of the split
	    FileSystem fs = file.getFileSystem(job);
	    fileIn = fs.open(split.getPath());
	    boolean skipFirstLine = false;
	    if (codec != null) {
	      //in = new LineReader(codec.createInputStream(fileIn), job);
	      end = Long.MAX_VALUE;
	    } else {
	      if (start != 0) {
	        skipFirstLine = true;
	        --start;
	        fileIn.seek(start);
	      }
	      
	      //in = new LineReader(fileIn, job);
	    }
	    if (skipFirstLine) {  // skip first line and re-establish "start".
	     /* start += in.readLine(new Text(), 0,
	                           (int)Math.min((long)Integer.MAX_VALUE, end - start));
	      */
	      start+=fileIn.read(new byte[PRISM_Settings.RIJNDAEL_BLOCK_SIZE], 0, (int)Math.min((long)Integer.MAX_VALUE, end - start));
	    }
	    marker=start;


	    //Erik
	    //System.out.println("Pos:"+this.pos);
	  }
	  
	  public boolean nextKeyValue() throws IOException {  
	    if (key == null) {
	      key = new LongWritable();
	    }
	    //key.set(pos);
	    
	    if (value == null) {
	      value = new BytesWritable();
	    }
	    int newSize = 0;
	    /*newSize = in.readLine(text_value, maxLineLength,
	                            Math.max((int)Math.min(Integer.MAX_VALUE, end-pos),
	                                     maxLineLength));
	      */
	      bytes_read = fileIn.read(buf);
	      
	      /*
	      if (newSize == 0) {
	        break;
	      }
	      */
	      if (bytes_read== -1) {
		        return false;//No more key because end of file has been reached
		  }
	      
	      //Erik
	      //System.out.println("Read "+bytes_read+"bytes");

	      //BROKEN pos:=fileIn.getPos();	      
	      marker+=1;
	      key.set(marker); // the marker follows the progression of the bytes read in the ciphered text
	      value = new BytesWritable(buf);//MODIFICATION

	      //Erik
	      //System.out.println("key: "+key.get());

	      //Erik
	      //System.out.println("bytes repr: "+ new Text(value.getBytes()));
	      
	    /*
	     *============ DEBUG===============================
	      //Null Pointer Exception on purpose for debugging
		    String s =null;
		    s.charAt(1);
	      */
	       
	     /* pos += newSize;
	      if (newSize < maxLineLength) {
	        break;
	      }
	        
	      */
	      
	      /*
	      // line too long. try again
	      LOG.info("Skipped line of size " + newSize + " at pos " + 
	               (pos - newSize));
	               
	      */
	      
	  //I Close while  }
	    /*
	    if (newSize == 0) {
	      key = null;
	      value = null;
	      return false;
	    } else {
	      return true;
	    }
	    
	    */
	      
	    return true; // this means the key and the value have been set properly
	  }

	  @Override
	  public LongWritable getCurrentKey() {
	    return this.key;
	  }

	  @Override
	  public BytesWritable getCurrentValue() {
	    return this.value;
	  }

	  /**
	   * Get the progress within the split
	   */
	  public float getProgress() {	    
	    if (start == end) {
	      return 0.0f;
	    } else { 
		//Erik
		float f=((float)(PRISM_Settings.RIJNDAEL_BLOCK_SIZE*marker - start))/((float)(end - start));
		return f;
	    }
	  }
	  
	  public synchronized void close() throws IOException {
	    if (fileIn != null) {
	      fileIn.close(); 
	    }
	  }
}
