import gnu.crypto.cipher.IBlockCipher;
import gnu.crypto.mac.IMac;
import gnu.crypto.mac.MacFactory;
import gnu.crypto.pad.IPad;
import gnu.crypto.pad.PadFactory;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;


public class Decrypt {
	   
	   public static void main(String[] args) throws Exception {
		   
		   	Date day = null; 
		   	
		   	if(args.length < 3){
		   		 System.out.println("Error Missing parameters");
				 System.out.println("Usage: Decrypt <input_ciphered_padded_text_file>  <ouput_deciphered_text_file> <Date :dd-MM-yyyy > [<Secret key K>]");
				 System.exit(1);
			 }
			 
			 if(args.length >= 4 && args[3].length() < PRISM_Settings.RJINDAEL_KEY_SIZE){
				 System.out.println("The Secret key you entered is too short");
				 System.out.println("The length of the key must be at least the length of the computed HMAC (here, 32 bytes).");
				 System.out.println("[for testing purpose] The secret key is an optional parameter, if ommited a default key string will be used.");
				 System.exit(1);
			 }
			 
			 String key_str;
			 
			 if(args.length >= 4 && args[3].length() >= PRISM_Settings.RJINDAEL_KEY_SIZE ){ // key size 32 bytes
				 key_str=args[3];
				 
			 }else { // Use the default key string
				 System.out.println("Using the default key string: 'allyourbasearebelongtoushahahaha' ");
				 // Creation of Kd
				 key_str = PRISM_Settings.RIJNDAEL_DEFAULT_KEY_STRING; 
				 /* For testing we use the fixed string:  "allyourbasearebelongtoushahahaha"
				 The length of the key must be at least the length of the computed hash (here, 32 bytes).
				 */
	   		}
			 
			 String day_str = args[2];
			 
			 RIJNDAEL_Cipher RCipher = new RIJNDAEL_Cipher(key_str,day_str);
			 
			 IBlockCipher cipher = RCipher.getCipher();
		     int bs = cipher.currentBlockSize();
		     
		     // Instanciation of the padding
		     // The word concatenated to the counter are transformed to fit the 32 bytes block size
		     IPad padding = PadFactory.getInstance("PKCS7");
		     padding.init(bs);
		     
		     int counter=0;
		     
		     try {
		    	 // I am using a FileInputStream for the input to work on bytes
		    	 File file_input = new File(args[0]);
		    	 FileInputStream in = new FileInputStream(file_input);
		    	 
		    	 // I am using a BufferedWriter for the output to write strings
		    	 File file_output = new File(args[1]);
		    	 BufferedWriter writer = null;
		    	 writer = new BufferedWriter(new FileWriter(file_output));
				 
		    	 byte[] encrypted = new byte[PRISM_Settings.RIJNDAEL_BLOCK_SIZE];

			 int i=0;				 


		    	 // We read the input, per block of 32 bytes
				 while (in.read(encrypted, 0, PRISM_Settings.RIJNDAEL_BLOCK_SIZE) == PRISM_Settings.RIJNDAEL_BLOCK_SIZE){
			
				     i++;
		 
					 byte[] plain_padded = new byte[encrypted.length];				 
					 // Decryption part
					 byte[] word_decrypted;

					 					
					 cipher.decryptBlock(encrypted, 0, plain_padded, 0);
					 
				     // We remove the padding from the decrypted word
				     int unpad = 0;
				     try{
				    	 unpad = padding.unpad(plain_padded, 0, plain_padded.length);
				     }catch (gnu.crypto.pad.WrongPaddingException ex){
					 // ex.printStackTrace();
				    	 
				    	 //System.exit(1);
				     }
				     
				     byte[] plain = new byte[plain_padded.length - unpad];
				     System.arraycopy(plain_padded, 0, plain, 0, plain.length);
				     // I split the decrypted byte into the bytes corresponding to the word, and the ones corresponding to the counter.
				     word_decrypted = new byte[plain.length - PRISM_Settings.COUNTER_LENGTH];
				     byte[] counter_decrypted = new byte[PRISM_Settings.COUNTER_LENGTH];
				     for (int k = 0; k < word_decrypted.length; k++)
				    	 word_decrypted[k] = plain[k];
				     for (int k = word_decrypted.length; k < plain.length; k++){
				    	 counter_decrypted[k - word_decrypted.length] = plain[k];
				     }
				     // Conversion of the counter's bytes into an int.
				     counter = counter_decrypted[0];
				     counter = (counter << 8) | (counter_decrypted[1] & 0xff);
					 

				     String plainText = new String(word_decrypted); // Transform the byte[] into string 


				     System.out.println(plainText);

				     
				     // We write the results into the output file
				     writer.write(plainText+ "  :  " + counter);
				     writer.newLine();
				     
				 }
				 in.close();
				 writer.close();
		     }
		     
		     catch (FileNotFoundException e) {
		    	 e.printStackTrace();
		    	 System.exit(1);
		    	 } 
		     catch (IOException e) {
		    	 e.printStackTrace();
		    	 System.exit(1);
		    	 }
	   }
}
