
import java.util.Random;

import gnu.crypto.cipher.IBlockCipher;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.math.BigInteger;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.ArrayList;

import java.util.Arrays;
import java.util.BitSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.filecache.DistributedCache;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.BytesWritable;

import org.apache.hadoop.io.ArrayWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.Reducer.Context;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;

import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

/*import org.bouncycastle.jce.ECNamedCurveTable;
import org.bouncycastle.math.ec.ECCurve;
import org.bouncycastle.math.ec.ECPoint;
import org.bouncycastle.math.ec.ECPoint.Fp;
*/


public class Search extends Configured implements Tool{
	//Adding necessary configuration files, without this files, the execution hangs at the instruction 
	//Job job = new Job(getConf()), this is certainly because to instantiate a new Job you need some info from those
	//configuration files
	static {
		Configuration.addDefaultResource("hdfs-default.xml");
		Configuration.addDefaultResource("hdfs-site.xml");
		Configuration.addDefaultResource("mapred-default.xml");
		Configuration.addDefaultResource("mapred-site.xml");
		Configuration.addDefaultResource("hadoop-policy.xml");
	}
	/**
	 * MAP operations: for each Ci, the cloud computes the candidate position Xi||Yi and fills its PIR matrix
	 * REDUCE operations :
	 */
	

	//static EC_ELGAMAL ecEGAMAL = new EC_ELGAMAL(PRISM_Settings.EC_ELGAMAL_DEFAULT_SECRET_KEY);
	
	//The following parameters can also be passed through command line or moved to PRISM_Settings class
	public final static String LOCAL_ALPHAS_VALUES="local_alphas_values.dat"; // Path to where the r1,r2,...rt values will be stored on the local computer
	public final static String HDFS_ALPHAS_VALUES="input/hdfs_aphas_values.dat";// Relative Path to where the r1,r2,...rt values will be stored on the HDFS
	
    //public final static String LOCAL_E_Pinf_VALUE="local_E_Pinf_value.dat"; // Path to where E(Pinf) value will be stored on the local computer
	//NB: E(Pinf) will be used as an initializator for the PIR row sums in the mappers
	//public final static String HDFS_E_Pinf_VALUE="input/hdfs_E_Pinf_value.dat";// Path to where the E(Pinf) value will be stored on the HDFS
	
        	public static class Map       
	extends Mapper<LongWritable, BytesWritable, Text, BytesArrayWritable > { //ArrayList<ECPoint.Fp[]>
		

	
		/**Mapper<KEYIN,VALUEIN,KEYOUT,VALUEOUT>
		 *In our case 
		  * KEYIN = block number in the ciphered text
		  * VALUEIN = ciphered block
		  * KEYOUT = input file name
		  * VALUEOUT = HashMap of PIR values: PIR1,PIR2.....,PIRq 
		  **/
       
		    long myTime = 0;

                BigInteger local_alphas[];
		//	ArrayList<ECPoint.Fp[]> local_alphas; // this is a local copy( belonging to this mapper) of the {a1,a2,....at} array of ciphertexts
		//ECPoint.Fp[] E_Pinf;
		//NB: E(Pinf) will be used as an initializator for the PIR row sums in the mappers
		//ArrayList<BitSet> Mlist;
                ArrayList<byte[]> Mlist2;
		

		int t = 0;// number of rows
		int q = 0; //Number of rounds
		int X_Y_LENGTH = 0; //number of bits to consider in ciphered bloc, in order to compute candidates positions
		
		/**
		   * Called once at the beginning of the task.
		   * Operations performed in this method :
		   * 1) retrieve parameters t and q, loading of alphas values array and E(Pinf) 
		   * 2) creation of PIR matrix M initially filled with zeros
		 */
		@Override
		protected void setup(Context context
        ) throws IOException, InterruptedException{
			
			System.out.println("Executing Mapper's setup method");
			
			/*********DEBUG********/
			//String s=null;
			//s.charAt(0);
			/**********************/
			
			boolean loaded_alphas_values=false;
			
			
			Configuration conf= context.getConfiguration();
			int t_exp = conf.getInt("t_EXP", PRISM_Settings.DEFAULT_EXP_FOR_NUMBER_OF_ROWS); // get the parameter named t_EXP, if none defined, take
																								// the default value specified in PRISM_Settings
			t= (int)Math.pow(2, t_exp);
			q=conf.getInt("Q", PRISM_Settings.DEFAULT_NUMBER_OF_ROUNDS);
			X_Y_LENGTH = 2*(int)PRISM_Utils.log2(t); //== 2*log2(t) length of the concatenation Xi||Yi
			
			System.out.println("t="+t+ " "+"q="+q + " " + "X_Y_LENGTH="+X_Y_LENGTH);
			/****DEBUG*********/
			/*String str=null;
			str.charAt(0);
			*/
			
			local_alphas = new BigInteger[t];

			// loading aphas(a1,a2,....at) values and E(Pinf) value
			try {
			    //			      String alphas_VALUES_CacheName = new Path(HDFS_ALPHAS_VALUES).getName();

			      //			      String E_Pinf_CacheName = new Path(HDFS_E_Pinf_VALUE).getName();
			      
			      //Path [] cacheFiles = DistributedCache.getLocalCacheFiles(context.getConfiguration());

			     		     
			      System.out.println("Opening Alpha Values in "+HDFS_ALPHAS_VALUES);
			     
Object obj = load_Cached_Value(HDFS_ALPHAS_VALUES); //Retrieve the ciphertexts {a1,a2,........at}
				        	if(obj instanceof BigInteger[]){
				        		BigInteger[] encoded=(BigInteger[])obj;
				        		
				        		//for(byte[][] b:encoded_Pts)
							for (int i=0;i<encoded.length;i++)
{
    
    this.local_alphas[i]=encoded[i];
    //System.out.println("Alpha No. "+i+": "+local_alphas[i]);
				        		}
				        		loaded_alphas_values = true;
				        	}
				        	
				          }

 catch (Exception ioe) {
			      System.err.println("IOException reading from distributed cache");
			      System.err.println(ioe.toString());
			    }


			/*



 if (null != cacheFiles && cacheFiles.length > 0) {
			    	  System.out.println("Cache files NOT NULL");
			    	  
			        for (Path cachePath : cacheFiles)
			        {
				        	
				          if (cachePath.getName().equals(alphas_VALUES_CacheName)) {
				        	  System.out.println("Loading alphas from HDFS");
				        	Object obj = load_Cached_Value(cachePath); //Retrieve the ciphertexts {a1,a2,........at}
				        	if(obj instanceof BigInteger[]){
				        		BigInteger[] encoded=(BigInteger[])obj;
				        		
				        		//for(byte[][] b:encoded_Pts)
							for (int i=0;i<encoded.length;i++)
{
    
    this.local_alphas[i]=encoded[i];
    //System.out.println("Alpha No. "+i+": "+local_alphas[i]);
				        		}
				        		loaded_alphas_values = true;
				        	}
				        	
				          }
				          
				          
				          if(loaded_alphas_values){
				        	  break;
				          }
				      }
			      }else{
			    	  System.out.println("cache Files not LOADED!");
			      }
			*/
			    
			    
			    //Creating list of q PIR matrix M, ==> one matrix for each round
			    //Mlist = new ArrayList<BitSet>();
			    Mlist2 = new ArrayList<byte[]>();
			    for (int i=0;i<q;i++){
				if (t==2) {
				    Mlist2.add(new byte[1]);
				}
				else {
				    Mlist2.add(new byte[(t*t/8)]);
				}
				
					//Mlist.add(new BitSet(t*t)); // Java by default fills the BitSet with false=zero at creation time
			    }
			    
			    myTime = System.currentTimeMillis();
			    //System.out.println("End of setup at: "+myTime);			     
		}
		/**
		 * Read the file identified by 'cachePath' and return the Object that object that was serialized in this file
		 * @param cachePath
		 * @return original Object that was serialized in the Distributed Cache
		 */
		public Object load_Cached_Value(String cachePath){
			FileInputStream fis = null;
			ObjectInputStream in = null;
			Object obj=null;
			
			try{
				fis = new FileInputStream(cachePath);
				in = new ObjectInputStream(fis);
				obj=in.readObject();
				in.close();
			}
			catch(IOException ex){
				ex.printStackTrace();
			} catch (ClassNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			return obj;
		}
		

		//Erik

		private boolean found = false;

		/**
		   * Called once for each key/value pair in the input split. 
		   * <KEYIN,VALUE_IN, KEY_OUT,VALUE_OUT> LongWritable, BytesWritable, LongWritable, BytesWritable
		*/
		public void map(LongWritable key, BytesWritable value, Context context)  
		throws IOException, InterruptedException {

if (key.get()%3==0) {
    if ((new String(value.getBytes()).compareTo("sdlfkj"))==0) {
	found = true;
    }
	        }
		}
		


protected void cleanup(Context context
                                         ) throws IOException, InterruptedException {

		    BytesWritable[] v= new  BytesWritable[1];
		    if (found) {
			v[0]= new BytesWritable((new String("found")).getBytes());}
		    else {
			v[0]= new BytesWritable((new String("not found")).getBytes());
		    }
		    BytesArrayWritable values = new BytesArrayWritable(v);
		    context.write(new Text(new String("found or not")), values); 
			 myTime = System.currentTimeMillis() - myTime;
			 System.out.println("Clean up finished, Map took : "+myTime);	    
}

	}
          
	public static class Reduce       
	extends Reducer<Text, BytesArrayWritable, Text, BytesArrayWritable > {
		
		/**
		  *Reducer <KEYIN,VALUEIN,KEYOUT,VALUEOUT>
		  *KEY_IN = Filename 
		  * VALUE_IN = {PIR1,PIR2,........,PIRq}
		  * KEY_OUT = Filename
		  * VALUE_OUT = {s_PIR1,s_PIR2,.....,sPIRq} --> list of ECPoint.Fp[]
		  * 
		 */
		
		int q=0; //number of rounds
		
		/**
		   * Called once at the start of the task.
		*/
		protected void setup(Context context ) throws IOException, InterruptedException {
			q=context.getConfiguration().getInt("Q", PRISM_Settings.DEFAULT_NUMBER_OF_ROUNDS);
			System.out.println("q="+q);
//			/****DEBUG*********/
//			String str=null;
//			str.charAt(0);
			
		}

		public void reduce(Text key, Iterable<BytesArrayWritable> values,     
		Context context) throws IOException, InterruptedException {     
			/*Decode the BytesArrayWritable and build an ArrayList<ECPoint.Fp[]>*/
		        BigInteger[] first_list; 
		
			Iterator<BytesArrayWritable> it = values.iterator();
		
			
			System.out.println("REDUCE!");
			
			if(it.hasNext()){
			    BytesArrayWritable first =(BytesArrayWritable) values.iterator().next(); 		

			first_list = Search.decodeBytesArray(first);
			//System.out.println("Received:");
			//printRow(first_list);
			
			while (it.hasNext()) {  // 1 <i < #mappers
				BigInteger[] PIRlist = Search.decodeBytesArray(it.next());
				for(int j=0;j<q;j++){ // 1 < j < q
				    first_list[j]  = first_list[j].add(PIRlist[j]);
				      
				}
			}
			
			//System.out.println("Send back:");
			//printRow(first_list);

		        BytesArrayWritable result = Search.encodeCipherTextArrayList(first_list);
			//System.out.println();

			context.write(key,result); // CLOUD OVER ^_^
			System.out.println("Reduce finished!");
			} else System.out.println("Empty Reduce?!");
		}          
	}
	

    public static BigInteger[] decodeBytesArray (BytesArrayWritable list) {
	BytesWritable[] by = list.getAll();

	BigInteger[] output = new BigInteger[by.length];
	for (int i=0;i<by.length;i++) {
	    output[i] = new BigInteger(by[i].getBytes());
	}
	return output;
    }


    public static void printRow (BigInteger[] sums) {
	for (int  i=0;i<sums.length;i++) {
	    System.out.println("Round "+i+", Sum="+sums[i]);
	}

    }

	/**
	 * Encode an arrayList of Elliptic CipherTexts
	 * @param list arrayList of Elliptic CipherTexts
	 * @return Array of BytesWritable, in this array each couple of BytesWritable represent the encoding of a cipherText
	 * therefore the size of this array is twice the size of the cipherText arrayList
	 */
	public static BytesArrayWritable encodeCipherTextArrayList(BigInteger [] source){
	 
	    //System.out.println("INPUT:");
	    //printRow(source);
	    
	    BytesWritable[] encoded_values = new BytesWritable[source.length];

	    for (int i=0;i<source.length;i++) {
		encoded_values[i] = new BytesWritable(source[i].toByteArray());
	    }

            BytesArrayWritable encoded_list=new BytesArrayWritable(encoded_values);

	    //System.out.println("OUTPUT");
	    //BigInteger[] x = Search.decodeBytesArray(encoded_list);
	    //printRow(x);


	    return encoded_list;
	}
	

	/**
	 * Returns the candidate position in the PIR matrix M for the given ciphered bloc 
	 * @param encrypted : ciphered bloc
	 * @param X_Y_LENGTH : number of bits to consider in ciphered bloc, in order to compute candidates positions,NB:this number should be a multiple of 2
	 * @return : candidate position [xi,yi] in PIR matrix
	 * NB: SHOULD we return long an not int values because ? the input bytes can be a big array of bytes whose value does not fit into an int...
	 */
	public static int[] get_candidate_x_y(byte[] encrypted,int X_Y_LENGTH){
		
		byte[] X;
		byte[] Y;
		
		BitSet X_Y = PRISM_Utils.getBits(encrypted, X_Y_LENGTH); // first 2*lof2(t) bits of the encrypted bloc
		
		BitSet b_x=X_Y.get(0, X_Y_LENGTH/2);
		BitSet b_y=X_Y.get(X_Y_LENGTH/2,X_Y_LENGTH);
		
		X=PRISM_Utils.toByteArray(b_x);
		Y=PRISM_Utils.toByteArray(b_y);
		
		String s_x = new BigInteger(X).toString();
		String s_y = new BigInteger(Y).toString();
		
		int x = Integer.valueOf(s_x); //bytes to int conversion
		int y = Integer.valueOf(s_y);
		
		//Erik
		//System.out.println("Candidates position");
		//System.out.println("x= "+x);
		//System.out.println("y= "+y);
		
		return new int[]{x,y};
	}
	
	/**
	 * Return the position in a (t*t)java BitSet [b(0,0),b(0,1),b(0,2).....b(0,t-1)
	 * 										 b(1,0),b(1,1),b(1,2).....b(1,t-1)
	 * 										 ----------------------------------
	 * 										 ----------------------------------								
	 * 										 b(t,0),b(1,1),b(1,2).....b(t-1,t-1) ] 
	 * corresponding to a tuple of coordinates (x,y)
	 * @param t : number of rows of the PIR matrix
	 * @param x x coordinate
	 * @param y y coordinate
	 * @return the position x*(t) + y corresponding to (x,y) in the bit set
	 */
	public static int getPosition(int t,int x, int y){
		return x*(t) + y;
	}
	/**
	 * 
	 * Transfer the local file where the alphas values ( HashMap ) are stored to the 
	 * Distributed cache of the HDFS so that they will be available to all mappers
	 * @param conf : Job configuration
	 * @throws IOException
	 */
	void send_to_Cloud(Configuration conf,String local_filename_path, String dest_filename_path) throws IOException {
		


	    FileSystem fs = FileSystem.get(conf);
	    Path hdfs_Path = new Path(dest_filename_path);

	    // upload the file to hdfs. Overwrite any existing copy.
	    //copyFromLocalFile(boolean delSrc, boolean overwrite, Path src, Path dst) 
	    fs.copyFromLocalFile(false, true, new Path(local_filename_path),
	        hdfs_Path);


	    System.out.println("HERE -- Local: "+local_filename_path+", HDFS: "+ hdfs_Path);
	    
	    DistributedCache.addCacheFile(hdfs_Path.toUri(), conf);
	    
	}
      
	public int run(String [] new_args) throws Exception {
		
		int q=0;
		int exp=0;
		
		String q_str = new_args[new_args.length -4]; //#rounds q
		String exp_str=new_args[new_args.length -3]; // value of t's exponent, remember t= 2^e
		
		try{
			q=Integer.parseInt(q_str);
			exp=Integer.parseInt(exp_str);
		}catch(NumberFormatException e){
			System.err.println("Error q and t must me integer values! Aborting!");
			System.exit(1);
		}
		
		String input_file = new_args[new_args.length-2];
		String output_file = new_args[new_args.length -1];
		
//		System.out.println("input_file:"+input_file);
//		System.out.println("output_file:"+output_file);
		
		System.out.println("Executing run method");
		/* DEBUG*/
//		Configuration c = getConf();
//		System.out.println("configuration:"+c.toString());
		
		Job job = new Job(getConf());
		
		Configuration conf = job.getConfiguration();
		//Configuration conf = getConf();
		Path p = new Path(input_file); //input file
		conf.set("INPUT_FILE", p.getName()); // the getName method returns the last component of the path, after the last separator
		conf.setInt("Q", q); //value for q
		conf.setInt("t_EXP", exp); //value for t's exponent
		
		System.out.println("Sending alphas to cloud");
		this.send_to_Cloud(conf,LOCAL_ALPHAS_VALUES,HDFS_ALPHAS_VALUES);
		
		//System.out.println("Sending E(Pinf) to cloud");
		//this.send_to_Cloud(conf, LOCAL_E_Pinf_VALUE, HDFS_E_Pinf_VALUE);
		
		System.out.println("Job Settings...");
		job.setJarByClass(Search.class);          
		job.setJobName("PRISM_Search");
		
				  
		job.setMapOutputKeyClass(Text.class);
		job.setMapOutputValueClass(BytesArrayWritable.class);
		
		//This is for the reduce job
		job.setOutputKeyClass(Text.class);
		job.setOutputValueClass(BytesArrayWritable.class);          
		
				  
		job.setMapperClass(Map.class); //THIS IS NOT TRUE, Map extends org.apache.hadoop.mapreduce.Mapper instead         
		//job.setCombinerClass(Reduce.class); 
		job.setReducerClass(Reduce.class);         
				  
		job.setInputFormatClass(BytesInputFormat.class);
		job.setOutputFormatClass(SearchOutputFormat.class);
		
		FileInputFormat.setInputPaths(job, new Path(input_file));        
		FileOutputFormat.setOutputPath(job, new Path(output_file));        
		
		System.out.println("Waiting for completion");
		long t = System.currentTimeMillis();
		boolean success = job.waitForCompletion(true);
		System.out.println("WALL CLOCK: "+((System.currentTimeMillis()-t)/1000));
		return success ? 0 : 1;     
	}          
    
	/**
	 * Gets the day from the input file path
	 * @param Inputfile_path
	 * @return the date in the form DD-MM-YYYY
	 */
	static private String getDay(String Inputfile_path){

		String date = "";
		String[] fields = Inputfile_path.split("/");
		String basename = fields[fields.length-1]; // last field
		System.out.println(basename);
		String[] date_fields = basename.split("-");
		date+=date_fields[0]+"-"+date_fields[1]+"-"+date_fields[2];
		return date;
	}
	
	public static void main(String[] args) throws Exception {
		
		if(args.length<8){ //If I don't have 8 arguments
			System.out.println("Usage: hadoop jar <jar_file> -libjars lib1,lib2,...libn Search <word_to_search> <input_file> <output_file> <exponent e used to compute t=2^e > <number of rounds> < RIJNDAEL symmetric encryption key > ");
			// <input_file> is the name of the file in the HDFS
			System.exit(1);
		}
		
		
		
		int t_exp=0;
		int t=0;//number or rows
		
		int len=args.length;
		String secret_key="";//Secret key for RIJNDAEL cipher
		String word=null;//word we are searching
		String input_file=null;
		String output_file=null;
		int libs_marker=0; // I use this as a marker to identify the index of the last external library
		
	
		
		// I parse the arguments starting from the last one, from right to left
		
		secret_key= args[len-1];
		String q_str=args[len-2]; //number or rounds
		String exp_str=args[len-3]; //exponent
		output_file=args[len -4]; //e.g : -o<output_search.txt>
		input_file=args[len - 5];
		word=args[len - 6];
		word=word.trim();
		libs_marker = args.length -6 ;
		
		System.out.println("word:"+word);
		
		
		try{
			t_exp=Integer.parseInt(exp_str);
		}catch(NumberFormatException e){
			System.err.println("t must me integer value! Aborting!");
			System.exit(1);
		}
		t=(int)Math.pow(2, t_exp);
		
		String day_str=getDay(input_file);//get the date from the input file name in the form DD-MM-YYYY
		
		
		Random rand = new SecureRandom();

		//SAMPLE
		//BigInteger p = new BigInteger("83");
		//BigInteger N = new BigInteger("10");
		//BigInteger b = new BigInteger("33");
		

		BigInteger m = new BigInteger("400");
		//BigInteger p = BigInteger.probablePrime(m.intValue(), rand);

		BigInteger p = new BigInteger("1564820277778863089963405534144523812327502518356860558287169907464574863492764499627171260275816613599242003651320891327");

	
		//N = HOW MANY MAPPERS
		BigInteger N = new BigInteger("23");
		BigInteger sqrt_n = new BigInteger((new Long(t)).toString());
		//BigInteger b = (new BigInteger(m.intValue(), rand)).mod(p);
		BigInteger b = new BigInteger("1298551010326079989062112961798042817861449001333409542125059506044661625781468711882345844134701566170769791452428186529");
		BigInteger b_inverse = b.modInverse(p);
		
		System.out.println("t: "+sqrt_n+", N: "+N);
		BigInteger N_to_r = N; 
		BigInteger N_to_l = BigInteger.ONE; 
		BigInteger e_limit = p.divide(sqrt_n.multiply(N.subtract(BigInteger.ONE)));
		BigInteger a_limit = e_limit.divide(N_to_r);    

		//System.out.println("E Limit: "+e_limit+", A Limit: "+a_limit);
	
		RIJNDAEL_Cipher RIJ = null;
		
		if(secret_key.equals("DEFAULT")){
			RIJ= new RIJNDAEL_Cipher(day_str); //This will use the default key string for our RIJNDAEL Cipher
		}else{
			RIJ= new RIJNDAEL_Cipher(secret_key,day_str);
		}
		
		IBlockCipher rij_cipher = RIJ.getCipher();
		int blocksize = rij_cipher.currentBlockSize();
		
		byte[] plain_bytes= PRISM_Utils.concatenate(word.getBytes(), 1);
		byte[] padded_bytes = RIJNDAEL_Cipher.padBytes(plain_bytes, blocksize);
		byte[] encrypted = new byte[padded_bytes.length];
		int X_Y_LENGTH = 2*(int)PRISM_Utils.log2(t); //== 2*log2(t) length of the concatenation Xi||Yi
		
		
		BigInteger alphas[] = new BigInteger[t];
		BigInteger a[] = new BigInteger[t];

		
		
		// Encryption
		rij_cipher.encryptBlock(padded_bytes, 0, encrypted, 0);
		
		int [] x_y = get_candidate_x_y(encrypted,X_Y_LENGTH);
		int x = x_y[0];
		int y = x_y[1];
		
		System.out.println("Candidate position is: "+x+", "+y);
		
		//U computes (t) ciphertexts {a1,a2,.....at}
		
    for (int i=0;i<t;i++) {
	a[i] = (new BigInteger(m.intValue(), rand)).mod(a_limit);
	//a[i] = (new BigInteger(20, rand)).mod(a_limit);
	if (i==y) {

	    alphas[i] = (N_to_l.add(a[i].multiply(N_to_r))).multiply(b);

	}
	else {

	    alphas[i] = (a[i].multiply(N_to_r)).multiply(b);

	}
	//System.out.println("Alpha No. "+i+": "+alphas[i]);
    }


		System.out.println("Serializing alpha values");
	      
		
		//First we serialize the aphas array into a local file that we will send later on to the cloud using distributed cache
		try{
			FileOutputStream fos1 = new FileOutputStream(LOCAL_ALPHAS_VALUES);
			ObjectOutputStream out1 = new ObjectOutputStream(fos1);
			
			out1.writeObject(alphas);
			fos1.close();
			out1.close();
			
			
		}
		catch(IOException ex){
			ex.printStackTrace();
			System.exit(1);
		}
		
		String[] new_args= new String[libs_marker +4 ];
		System.arraycopy(args, 0, new_args, 0, libs_marker);
		new_args[new_args.length -4]=q_str; //#rounds q
		new_args[new_args.length -3]=exp_str; // value of t's exponent, remember t= 2^e
		new_args[new_args.length - 2] = input_file;
		new_args[new_args.length - 1] = output_file;
		
		System.out.println("Running Search algorithm");
		int ret = ToolRunner.run(new Configuration(),new Search(), new_args); // Pinf will be used to initialize the PIR row sum
		System.exit(ret);          
				  
	}
}
