
import java.util.BitSet;
import java.util.Calendar;

import org.bouncycastle.math.ec.ECPoint;

import gnu.crypto.pad.IPad;
import gnu.crypto.pad.PadFactory;

/**
 * This Class regroups a set of methods(tools) useful for our project such as : 
 * concatenate(byte[] b, int i) which concatenates a byte array with an integer
 * 
 *
 */
public class PRISM_Utils {
	   /**
	    * concatenate the bytes of b with the counter i written on 2 bytes
	    * @param b bloc of bytes
	    * @param i integer to concatenate to b
	    * @return the resulting byte[] concatenation
	    * NB: in this method we assume that the integer stands on 2 bytes !
	    */
	   public final static  byte[] concatenate(byte[] b,int i){ 
		   byte[] concat;
		   concat = new byte[b.length + PRISM_Settings.COUNTER_LENGTH];
		   System.arraycopy(b, 0, concat, 0, b.length);
		   // The counter is put at the end of the word, on 2 bytes
		   concat[b.length+1] = (byte) (0xff & i);
		   concat[b.length] = (byte) (((0xff << 8) & i) >> 8);
		   return concat;
	   }
	   

    static void setBit(byte[] data, int pos, int val) {
	int posByte = pos/8; 
	int posBit = pos%8;
	byte oldByte = data[posByte];
	oldByte = (byte) (((0xFF7F>>posBit) & oldByte) & 0x00FF);
	byte newByte = (byte) ((val<<(8-(posBit+1))) | oldByte);
	data[posByte] = newByte;
    }

    static int getBit(byte[] data, int pos) {
	int posByte = pos/8; 
	int posBit = pos%8;
	byte valByte = data[posByte];
	int valInt = valByte>>(8-(posBit+1)) & 0x0001;
	return valInt;
    }






	   /**
	    * Returns a bitset of size num_bits containing the first num_bits bits in the byte array b.
	    * @param b input byte array
	    * @param num_bits number of leading bits of byte array b to take 
	    * @return BitSet of num_bits containing the first num_bits bits in the byte array b
	    */
	   public static BitSet getBits(byte[] b, int num_bits) {
	     BitSet bits = new BitSet();
	     for (int i = 0; i < num_bits  ; i++) {
	       if ((b[b.length - i / 8 - 1] & (1 << (i % 8))) > 0) {
	         bits.set(i);
	       }
	     }
	     return bits;
	   }
	   
	 /**
	  * Returns a byte array of at least length 1
	 The most significant bit in the result is guaranteed not to be a 1
	 (since BitSet does not support sign extension).
	 The byte-ordering of the result is big-endian which means the most significant bit is in element 0.
	 The bit at index 0 of the bit set is assumed to be the least significant bit.
	  * 
	  * @param bs the input BitSet
	  * @return Returns a byte array of at least length 1
	  */
	 public static byte[] toByteArray(BitSet bs) {
	    byte[] bytes = new byte[bs.length()/8+1];
	    for (int i=0; i<bs.length(); i++) {
	        if (bs.get(i)) {
	            bytes[bytes.length-i/8-1] |= 1<<(i%8);
	        }
	    }
	    return bytes;
	 }
	   
	    // log2:  Logarithm base 2
		public static double log2(double d) {
		      return Math.log(d)/Math.log(2.0);
		}
		
		/**
		 * Return the current date as a string
		 * @return the current date in the format : "dd/MM/yyyy"
		 */
		public static String getCurrentDate(){
			// The date is obtained and put in a byte[]
			 Calendar cal = Calendar.getInstance();
			 int d = cal.get(Calendar.DATE);
			 int month = cal.get(Calendar.MONTH) + 1;
			 int year = cal.get(Calendar.YEAR);
			 
			 String day_str = d+"-"+month+"-"+year;
			 
			/* System.out.println("Day: " + day_d);
			 System.out.println("Month: " + month);
			 System.out.println("Year: " + year);
			 */
		     return day_str;
		}

}
