#!/bin/bash
PRISM_HOME=`pwd` #this is the root directory of the PRISM_Project

HDFS=hdfs://iwak.eurecom.fr:9000  #Name node of our cluster
#HDFS=hdfs://localhost:9000      #local namenode [for testing purpose]

function usage {
  echo "Usage : $0 <user_name> <local_input_file> [ <secret key K> ]"
}

#We check if the hadoop executable is in the PATH variable, otherwise we abort the scrip,
# A solution could be to use the find command to get the folder where it is located, but for the moment for simplicity reasons,  let us just include hadoop executable in the PATH

type -P hadoop &>/dev/null || { echo " ### This script requires hadoop to be installed and the hadoop executable to be in the PATH variable, please install hadoop if not already installed and include the bin folder of hadoop setup in the PATH . Aborting. ###" >&2; exit; } # or 'exit' 1 if needed

if [ $# -lt 2 ]
then
    echo "### missing parameter(s) ###"
    usage #call the usage() function
    exit
fi

user_name=$1
input_file_path=$2 # input filepath e.g : /mydir/logs/logfile_xyz.txt
secret_key=''

date_d=`date '+%d-%m-%Y'` # today's date in the format : "dd-mm-yyyy"
echo "date is : " $date_d

if [ $# -ge 3 ] #if we have at least 3 parameters
then
  secret_key=$3
else
  echo "*** The default secret key will be used for encryption ***"
  secret_key='DEFAULT'
fi

bash compile_sources.sh # run the compile_sources.sh script

input_file=`echo $input_file_path | awk -F '/' '{print $NF }'` #we take only the filename which is the last element of the path
local_output_fname=$date_d-$input_file #the encrypted output_file name that will be written on the local disk e.g : "10-01-2011-logfile_xyz.txt"

echo "*** Encryption of the input_file ***"
if [ "$secret_key" == "DEFAULT" ]
then
  java -Xmx1500M -classpath .:$PRISM_HOME/classes/:$PRISM_HOME/src/:$PRISM_HOME/libs/*:$PRISM_HOME/libs/gnu-crypto-2.0.1-bin/lib/* RIJNDAEL_Cipher $input_file_path $local_output_fname
else
  java -classpath .:$PRISM_HOME/classes/:$PRISM_HOME/src/:$PRISM_HOME/libs/*:$PRISM_HOME/libs/gnu-crypto-2.0.1-bin/lib/* RIJNDAEL_Cipher $input_file_path $local_output_fname $secret_key
fi

status=$? #return code pf previous java command 

if [ $status -ne 0 ]
then
  echo "### Error while Encrypting file.Aborting ###"
  exit 1
fi

echo "*** Starting Upload into HDFS, the encrypted file will be copied to the Enc_Input folder in your HDFS user account *** "

#hadoop dfsadmin -safemode leave > /dev/null  # quit the safemode

hadoop fs -D dfs.block.size=100663296 -copyFromLocal $local_output_fname $HDFS/user/$user_name/Enc_Input/$local_output_fname
status=$? #return code pf previous command

if [ $status -ne 0 ]
then
  echo "### HDFS Error occured while uploading file. Aborting ###"
  exit 1
fi

echo "Removing temporary local file : $local_output_fname"
rm $local_output_fname

echo "*** Done! Please browse your to home folder in the HDFS to see if the file has be correctly uploaded ***"
