#!/bin/bash
PRISM_HOME=`pwd` #this is the root directory of the PRISM_Project, which should also be the directory containing this script
JAVA_CLASS=Decrypt

#HDFS=hdfs://iwak.eurecom.fr:9000  #Name node of our cluster
HDFS=hdfs://localhost:9000      #local namenode [for testing purpose]

function usage {
  echo "Usage : $0 <user_name> <input_ciphered_padded_text_file (on HDFS)>  [ <Secret key K> ] "
}

type -P hadoop &>/dev/null || { echo " ### This script requires hadoop to be installed and the hadoop executable to be in the PATH variable, please install hadoop if not already installed and include the bin folder of hadoop setup in the PATH . Aborting. ###" >&2; exit; } # or 'exit' 1 if needed 

if [ $# -lt 2 ]
then
    echo "### missing parameter(s) ###"
    usage #call the usage() function
    exit
fi

user_name=$1
input_file_path=$2
secret_key=''

input_file=`echo $input_file_path | awk -F '/' '{print $NF }'` #we take only the filename which is the last element of the path
output_file=$input_file # the name of the decrypted output file will be the same as the file in the HDFS

day=`echo $input_file | cut -d "-" -f1,2,3` #e.g echo 16-01-2011-hello.txt | cut -d "-" -f1,2,3 --> 16-01-2011

echo "day recovered from input file name: "$day

local_input_fname=LOCAL_$input_file #the encrypted input_file name that will be downloaded from the HDFS and written on the local disk

if [ $# -ge 3 ] #if we have at least 3 parameters
then
  secret_key=$3
else
  echo "*** The default secret key will be used for decryption ***"
  secret_key='DEFAULT'
fi

bash compile_sources.sh # run the compile_sources.sh script

echo "*** Downloading input file from HDFS, the file will be stored on the current directory then will be removed after decryption*** "

#if the file is already present, we delete it and we download a newer version
if [ -e $local_input_fname ]
then
  echo "Local file : $local_input_fname already exist, it will be deleted before downloading new version"
  rm $local_input_fname
fi

#hadoop dfsadmin -safemode leave > /dev/null #  quit the safemode
Echo "Name"
echo $HDFS/$input_file_path

#copy the file from HDFS to local disk
hadoop fs -copyToLocal $HDFS/$input_file_path $local_input_fname
status=$? #return code of previous command

if [ $status -ne 0 ]
then
  echo "HDFS Error occured while downloading file.Aborting"
  exit 1
fi

echo -e "\n*** Decryption of the input_file ***"
if [ "$secret_key" == "DEFAULT" ]
then
  java -classpath .:$PRISM_HOME/classes/:$PRISM_HOME/src/:$PRISM_HOME/libs/*:$PRISM_HOME/libs/gnu-crypto-2.0.1-bin/lib/* $JAVA_CLASS $local_input_fname $output_file $day
else
  java -classpath .:$PRISM_HOME/classes/:$PRISM_HOME/src/:$PRISM_HOME/libs/*:$PRISM_HOME/libs/gnu-crypto-2.0.1-bin/lib/* $JAVA_CLASS $local_input_fname $output_file $day $secret_key
fi

status=$? #return code pf previous java command 

if [ $status -ne 0 ]
then
  echo "### Error while decrypting file.Aborting ###"
  exit 1
fi

echo "Removing temporary local file :$local_input_fname"
rm $local_input_fname

echo "*** Done! Please open output_file $output_file to see if HDFS input file $input_file has be correctly decrypted ***"
