
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.math.BigInteger;
import java.security.KeyPairGenerator;
import java.security.SecureRandom;
import java.security.Security;
import java.util.ArrayList;
import java.util.Arrays;

import org.apache.hadoop.fs.Path;
import org.bouncycastle.jce.ECNamedCurveTable;

import org.bouncycastle.jce.spec.ECParameterSpec;
import org.bouncycastle.math.ec.ECCurve;
import org.bouncycastle.math.ec.ECFieldElement;
import org.bouncycastle.math.ec.ECPoint;
import org.bouncycastle.math.ec.ECPoint.Fp;



/**NB: FOR NOE THE SECRET KEY IS SET MANUALLY TO 1024, this is just for DEBUGGING PURPOSE ,later on the secret key will be a parameter of the constructor*/
public class EC_ELGAMAL {
	
	ECParameterSpec ecSpec;
	ECCurve.Fp curve;
	ECPoint.Fp P; // the base point for the EC curve
	ECPoint.Fp Pinf; // the infinity point for the EC curve
	BigInteger n;//the order N of G, n is the smallest number such that nP = infinity point
	BigInteger q; // parameter of the finite field Fq
	byte[] curve_seed; //seed for generating curve parameters
	ECPoint.Fp public_key;
	private byte[] sk; // secret Key bytes
	private BigInteger secret_key; //secret Key
	SecureRandom rand; // Secure Pseudo Random number generator
	
	//We can also pass the elliptic curve type as a parameter
	public EC_ELGAMAL(Long secretKey){
		ecSpec = ECNamedCurveTable.getParameterSpec("prime192v1"); // Elliptic Curve prime192v1 size = 192 bits
		curve = (ECCurve.Fp)ecSpec.getCurve();
		P=(ECPoint.Fp)ecSpec.getG();
		q = curve.getQ();
		Pinf=(ECPoint.Fp)curve.getInfinity();
		curve_seed = ecSpec.getSeed();
		this.rand = new SecureRandom();
		this.rand.setSeed(this.curve_seed); // I initialize the Random numer generator with the same seed used for generating the curve
		
		n= ecSpec.getN();
		//this.secret_key = this.createPrivateKey(); // I don't use this because I want a user supplied secret key
		this.secret_key = BigInteger.valueOf(secretKey);
		this.public_key =(ECPoint.Fp) P.multiply(secret_key);
	}
	/**
	 * create the private key for this cipher
	 * @return BigIntger encapsulating the bytes of the secret key
	 * NB: I don't use this method anymore because I need a user defined secret key
	 */
	private BigInteger createPrivateKey(){
		int n_bit_length = this.n.bitLength();
		int priv_key_size = (int) (n_bit_length/8) - 1 ; /* The secret key belongs to Fq and is used in point multiplication
		 * to produce the public key, and we know that all elements
		 * in Fq are integers between 0 and q -1, and in particular,
		 * the scalars used for point multiplicaton should be between 0 and order_of(P) - 1 
		 * for simplicity we choose the secret key to be an order of byte
		 * smaller than the order n
		 **/
		sk = new byte[priv_key_size];
		rand.nextBytes(sk); // fill sk with random bytes ==> generate the secret key
		return new BigInteger(sk);
	}
	/**
	 * ECPoint.Fp 
	 * getPublicKEy()
	 * @return the public key for this cipher
	 */
	public ECPoint.Fp getPublicKEy(){
		return this.public_key;
	}
	
	/**
	 * ECParameterSpec 
	 * getECParamaterSpec()
	 * @return the parameter Spec for the elliptic curve associated with this cipher
	 */
	
	public ECParameterSpec getECParamaterSpec(){
		return ecSpec;
	}
	
	/**
	 * ECCurve.Fp 
	 * getCurve()
	 * @return the elliptic curve associated with this cipher
	 */
	public ECCurve.Fp getCurve(){
		return this.curve;
	}
	
	/**
	 * SecureRandom 
	 * getSecureRandom()
	 * @return the secure Random number generator associated with this cipher's elliptic Parameter Specification
	 */
	public SecureRandom getSecureRandom(){
		return this.rand;
	}
	
	/**
	 * ECPoint.Fp 
	 * getBasePoint()
	 * @return the base point of the elliptic curve associated with this Cipher
	 */
	public ECPoint.Fp getBasePoint(){
		return this.P;
	}
	
	/**
	 * ECPoint.Fp 
	 * getInfinityPoint()
	 * @return the infinity point of the elliptic curve associated with this Cipher
	 */
	public ECPoint.Fp getInfinityPoint(){
		return this.Pinf;
	}
	
	/**
	 * int getKeySize()
	 * @return Key size in bytes of the private key
	 */
	public int getKeySize(){
		return secret_key.bitLength();
	}
	
	/**
	 * Encrypts an elliptic curve point using the El Gamal encryption scheme
	 * @param X the point X is the mapping of plaintext x to a point on the curve
	 * @param r random number
	 * @param sk_size secret key size
	 * @return array of 2 ECPoints representing the encryption of input ECPoint
	 */
	public ECPoint.Fp[] encryptV2(ECPoint.Fp X, byte r[],int sk_size){ // the point X is the mapping of plaintext x to a point on the curve
		BigInteger R = new BigInteger(r);
		ECPoint.Fp u = (ECPoint.Fp)P.multiply(R); // rP
		ECPoint.Fp prod = (ECPoint.Fp)public_key.multiply(R); // R*Y, Y = pub key
		ECPoint.Fp v = (ECPoint.Fp)prod.add(X); // R*Y + X
		return new ECPoint.Fp[]{u,v};
	}
	
	/**
	 * ECPoint.Fp[] 
	 * encrypt(ECPoint.Fp X, long r)
	 * Encrypts an elliptic curve point using the El Gamal encryption scheme
	 * @param X the point X is the mapping of plaintext x to a point on the curve
	 * @param r random number
	 * @return array of 2 ECPoints representing the encryption of input ECPoint
	 */
	public ECPoint.Fp[] encrypt(ECPoint.Fp X, long r){ // the point X is the mapping of plaintext x to a point on the curve
		BigInteger R = BigInteger.valueOf(r);
		ECPoint.Fp u = (ECPoint.Fp)P.multiply(R); // rP
		ECPoint.Fp prod = (ECPoint.Fp)public_key.multiply(R); // r*Y, Y = pub key
		ECPoint.Fp v = (ECPoint.Fp)prod.add(X); // r*Y + X
		return new ECPoint.Fp[]{u,v}; //(rP,r*Y + X)
	}
	
	/**
	 * ECPoint.Fp decrypt(ECPoint.Fp[] C)
	 * Decrypts a ciphetext C = (u,v) according to the scheme : D(C) = v - sk.u
	 * @param C ciphertext = (u,v)
	 * @return plaintext D(C) = v - sk.u
	 */
	public ECPoint.Fp decrypt(ECPoint.Fp[] C){ // C = (u,v)
		ECPoint.Fp Q = (ECPoint.Fp)C[1].subtract((ECPoint.Fp)C[0].multiply(secret_key)); // Q = v - sk.u
/*		NB: for the decryption of Pinf we have: Decrypt(E(Pinf)) = v - sk.u = (r.sk.P + Pinf) - sk.r.P // Remember for any ECPoint X : X + Pinf = X
																	 = r.sk.P - sk.r.P	
																	 = null_ECPoint = Pinf i.e (X=null,Y=null) */	
		return Q;
	}
	
	/**
	 * Decodes an Encoded Elliptic Curve Ciphertext -> (byte[], byte[]) to the equivalent Elliptic Curve Ciphertext -> (ECPoint.Fp,ECPoint.Fp) 
	 * @param encoded Encoded Elliptic Curve Ciphertext -> (byte[], byte[]) 
	 * @return decoded Elliptic Curve Ciphertext -> (ECPoint.Fp,ECPoint.Fp)
	 */
	public ECPoint.Fp[] decodeCiphertext(byte[][] encoded){
		ECPoint.Fp u = (ECPoint.Fp)curve.decodePoint(encoded[0]);
		ECPoint.Fp v = (ECPoint.Fp)curve.decodePoint(encoded[1]);
		
		ECPoint.Fp[] decoded = new ECPoint.Fp[]{u,v};
		return decoded;
	}
	
	public void printPoint(String name, ECPoint.Fp Pt){
		if(Pt.equals(this.Pinf)){
			System.out.println(name+": this point is Infinity Point!! (X=null,Y=null)");
		}else{
			System.out.println(name +": ( "+ Pt.getX().toBigInteger() +","+ Pt.getY().toBigInteger() + ")");
		}
	} 
	public Object load_Cached_Value(String filename){
		FileInputStream fis = null;
		ObjectInputStream in = null;
		Object obj=null;
		
		try{
			fis = new FileInputStream(filename);
			in = new ObjectInputStream(fis);
			obj=in.readObject();
			//in.readFully(buffer);
			in.close();
		}
		catch(IOException ex){
			ex.printStackTrace();
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return obj;
	}
	public static void main(String[] args) throws Exception {
		
		Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
		EC_ELGAMAL ecEG = new EC_ELGAMAL(PRISM_Settings.EC_ELGAMAL_DEFAULT_SECRET_KEY);
		ECCurve.Fp curve = ecEG.getCurve();
		ECPoint.Fp P = ecEG.getBasePoint();
		ECPoint.Fp Pinf = ecEG.getInfinityPoint();
		SecureRandom rnd = ecEG.getSecureRandom();
		
		System.out.println("The finite field for this curve is Fq with q= "+ecEG.q.toString());
		int field_size = curve.getFieldSize(); //equivalent to  return q.bitLength();
		System.out.println("curve field size = "+field_size);
		System.out.println("size = "+ecEG.q.bitLength());
		System.out.println("Order of base point P : "+ecEG.n.toString() + " bit length : "+ecEG.n.bitLength());
		System.out.println("private key size : "+ecEG.getKeySize()*8 + " bits");
		
		// Mapping : plaintext --> point on curve
		/* x | P(x)
		 * --|----
		 * 0 |Pinf
		 * --|----
		 * 1 |P
		 * 
		 * e.g: E(0) = (r.P, r.pub_key+Pinf)
		 */
		
		// C1 = E(0) = (u1,v1) tuple of Points with 0 ==> Pinf 
		/*byte[] r1 = new byte[priv_key_size];
		rnd.nextBytes(r1); // fill r1 with a random number
		*/
		
		long r1 = Math.abs(rnd.nextLong()); // generate a random number
		//long r1 = 2;
		ECPoint.Fp[] Enc_0 = ecEG.encrypt(Pinf,r1);
		
		long r2 = Math.abs(rnd.nextLong());
		//long r2 = 3;
		ECPoint.Fp[] Enc_1 = ecEG.encrypt(P,r2);
		
		System.out.println();
		
		// Now I compute E(P + Pinf) == E(1 + 0) ==  E(1) == E(P)
		ECPoint.Fp P_plus_Pinf = (ECPoint.Fp)P.add(Pinf);
		ECPoint.Fp P_plus_2Pinf = (ECPoint.Fp)P_plus_Pinf.add(Pinf);
		
		ecEG.printPoint("   P    ", P);
		ecEG.printPoint("P + Pinf", P_plus_Pinf);
		ecEG.printPoint("P + 2.Pinf", P_plus_2Pinf);
		
		if(P.equals(P_plus_Pinf)){
			System.out.println("Yeah it's true that P + Pinf = P");
		}
		
		System.out.println(P_plus_2Pinf.equals(P)?"P + 2Pinf == Pinf":"P + 2Pinf != Pinf" );
		
		long r = r1 + r2 ;
		System.out.println("r == "+r);
		ECPoint.Fp[] result_r2 = ecEG.encrypt(P_plus_Pinf,r2);
		ECPoint.Fp[] result_r = ecEG.encrypt(P_plus_Pinf,r);
		System.out.println("Comparison : E(P + Pinf) and E(P)");
		
		ecEG.printPoint("U_(Enc(P + Pinf) with r)  ",result_r[0]);
		
		ECPoint.Fp r1_x_P = (ECPoint.Fp)P.multiply(BigInteger.valueOf(r1));
		ECPoint.Fp r2_x_P = (ECPoint.Fp)P.multiply(BigInteger.valueOf(r2));
		
		ECPoint.Fp u1_plus_u2 = (ECPoint.Fp)r1_x_P.add(r2_x_P); // this should be equal to (r1+r2)*P
		ecEG.printPoint("        r1.P + r2.P       ", u1_plus_u2);
		ecEG.printPoint("           U1 + U2        ", (ECPoint.Fp)Enc_0[0].add(Enc_1[0]));
		ecEG.printPoint("U_(Enc(P + Pinf) with r2) ",result_r2[0]);
		ecEG.printPoint("           U_Enc_1        ",Enc_1[0]);
		
		ecEG.printPoint("V_Enc(P + Pinf) with r2)  ",result_r2[1]);
		ecEG.printPoint("           V_Enc_1        ",Enc_1[1]);
		
		ecEG.printPoint("V_Enc(P + Pinf) with r)   ",result_r[1]);
		
		ecEG.printPoint("     V_Enc(0) +V_Enc(1)   ", (ECPoint.Fp)Enc_0[1].add(Enc_1[1]));
		
		//Decryption tests
		
		System.out.println("Decryption :");
		
		//ECPoint.Fp Dec_1 = (ECPoint.Fp)Enc_1[1].subtract((ECPoint.Fp)Enc_1[0].multiply(secret_key));
		ECPoint.Fp Dec_0 = ecEG.decrypt(Enc_0);
		System.out.println(Dec_0.equals(Pinf)?"Dec(E(0)) == Pinf" : "Dec(E(0)) != Pinf");
		ECPoint.Fp Dec_1 = ecEG.decrypt(Enc_1);
		System.out.println(Dec_1.equals(P)?"Dec(E(1)) == P" : "Dec(E(1)) != P");
		
		//ECPoint.Fp Dec_P_plus_Pinf = (ECPoint.Fp)result_r[1].subtract((ECPoint.Fp)result_r[0].multiply(secret_key));
		ECPoint.Fp Dec_P_plus_Pinf = ecEG.decrypt(result_r);
		System.out.println(Dec_P_plus_Pinf.equals(P)?"Dec(E(0+1)) == P == Dec(E(1))" : "Dec(E(P+Pinf)) != P");
		
		//Tests
		System.out.println("Other tests on Point multiplication");
		ECPoint.Fp P1 = (ECPoint.Fp)P.multiply(BigInteger.valueOf(1)); // P1 = 1.P == P
		ecEG.printPoint("    P   ",P1);
		ecEG.printPoint("P1 = 1.P",P1);
		
		ECPoint.Fp P2 = (ECPoint.Fp)P.multiply(BigInteger.valueOf(2)); // P2 == 2P
		
		ECPoint.Fp P3 = (ECPoint.Fp)P1.add(P2); // P3 = P1 + P2 = P + 2P == 3P
		
		ECPoint.Fp P4 = (ECPoint.Fp)P.multiply(BigInteger.valueOf(3)); // P4 == 3P
		// P3 must be equal to P4
		ecEG.printPoint("P3 = P1 + P2 = P + 2P == 3P",P3);
		ecEG.printPoint("           P4 == 3*P       ",P4);
		System.out.println(P3.equals(P4)?"P3==P4" : "P3!=P4");
		
		System.out.println("\nTest Encoding and decoding");
		
		byte [][] encoded_E_0 = new byte[2][]; // C = Enc(0) = (u,v)
		System.out.println("Encoding...");
		encoded_E_0[0] = Enc_0[0].getEncoded(); //u
		encoded_E_0[1] = Enc_0[1].getEncoded();//v
		
		//Decoding
		System.out.println("Decoding...");
		ECPoint.Fp u = (ECPoint.Fp)curve.decodePoint(encoded_E_0[0]);
		ECPoint.Fp v = (ECPoint.Fp)curve.decodePoint(encoded_E_0[1]);
		
		ecEG.printPoint("u_E(0)",Enc_0[0]);
		ecEG.printPoint("v_E(0)",Enc_0[1]);
		System.out.println("Decoded E(0)");
		ecEG.printPoint("u",u);
		ecEG.printPoint("v",v);
		
		ECPoint.Fp[] decoded = new ECPoint.Fp[2];
		decoded[0]=u;
		decoded[1]=v;
		System.out.println("Decryption of decoded ciphertext (u,v)");
		ECPoint.Fp Point = ecEG.decrypt(decoded);
		ecEG.printPoint("decrypted Point",Point);
		if(Point.equals(Pinf)){
			System.out.println("decrypted point = 0 = Pinf ==> The coding and decoding works fine!!");
		}
		
		System.out.println("Testing serialization of encoded Ciphetexts");
		ArrayList<byte[][]> list = new ArrayList<byte[][]>();
		
		byte [][] encoded_E_1 = new byte[2][]; // C = Enc(1) = (u',v')
		encoded_E_1[0] = Enc_1[0].getEncoded(); //u'
		encoded_E_1[1] = Enc_1[1].getEncoded();//v'
		System.out.println("encoded_E_1[0] "+Arrays.toString(encoded_E_1[0]));
		System.out.println("encoded_E_1[0]length: "+encoded_E_1[0].length);
		
		System.out.println("encoded_E_1[1] "+Arrays.toString(encoded_E_1[1]));
		System.out.println("encoded_E_1[1]length: "+encoded_E_1[1].length);
		
		list.add(encoded_E_0);
		list.add(encoded_E_1);
		
		System.out.println("Serializing E(0) and E(1)");
		try{
			FileOutputStream fos1 = new FileOutputStream("encodedPoints.ser");
			ObjectOutputStream out1 = new ObjectOutputStream(fos1);
			
			out1.writeObject(list);
			fos1.close();
			out1.close();
			
			
			FileOutputStream fos2 = new FileOutputStream("encoded_E_Pinf.ser");
			ObjectOutputStream out2 = new ObjectOutputStream(fos2);
			out2.writeObject(encoded_E_0); // Remember 0 maps to Pinf therefore E(0) = E(Pinf)
			fos2.close();
			out2.close();
			
		}
		catch(IOException ex){
			ex.printStackTrace();
			System.exit(1);
		}
		
		System.out.println("Deserializing encoded ArrayList( encoded_E(0); encoded_E(1) ) ");
		Object obj1 = ecEG.load_Cached_Value("encodedPoints.ser");
		ArrayList<byte[][]> points=null;
		if(obj1 instanceof ArrayList<?>){
			points = (ArrayList<byte[][]>)obj1;
		}
		System.out.println("ArrayList contains:"+points.size()+"elements");
		
		for(byte[][] b:points){
		ECPoint.Fp[] dec = ecEG.decodeCiphertext(b);
		ECPoint.Fp Pt= ecEG.decrypt(dec);
			if(Pt.equals(P)){
				System.out.println("This point is Base Point : P!!");
			}
		ecEG.printPoint("decoded Point", Pt);
		}
		
		System.out.println("Deserializing encoded E(Pinf)...");
		Object obj2 = ecEG.load_Cached_Value("encoded_E_Pinf.ser");
		byte[][] encoded_E_Pinf = null;
		if(obj2 instanceof Object){
			System.out.println("Bingo");
			encoded_E_Pinf = (byte[][])obj2;
		}
		
		System.out.println("encoded_E_Pinf[0] "+Arrays.toString(encoded_E_Pinf[0]));
		System.out.println("E(Pinf)[0] length: "+encoded_E_Pinf[0].length);
		
		ECPoint.Fp[] decoded_E_Pinf=ecEG.decodeCiphertext(encoded_E_Pinf);
		ecEG.printPoint("decoded[0]", decoded_E_Pinf[0]);
		ecEG.printPoint("decoded[1]", decoded_E_Pinf[1]);
		ECPoint.Fp Pub_key_x_r=(ECPoint.Fp)ecEG.getPublicKEy().multiply(BigInteger.valueOf(r1));
		ecEG.printPoint("   r1*sk*P    ", Pub_key_x_r);
		// We know : E(Pinf) = (r*P,r*sk*P+ Pinf) = (r*P,r*sk*P) = (r*P, r*Pub_Key), therefore E_Pinf[1] should be == r*Pub_Key
		System.out.println(decoded_E_Pinf[1].equals(Pub_key_x_r)?"E_Pinf[1]==r1*sk*P":"E_Pinf[1] != r1*sk*P");
		ECPoint.Fp Pt2 =ecEG.decrypt(decoded_E_Pinf);
		ecEG.printPoint("decoded Point", Pt2);
	}

}
