import java.math.BigInteger;
import java.security.Key;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.SecureRandom;
import java.security.Security;
import java.util.Arrays;

import javax.crypto.Cipher;
public class ELGAMAL_Cipher {
	//NB: With El Gamal, the block of ciphertext produced is twice the size of the key—unlike RSA
	Cipher cipher;
	SecureRandom random;
	Key pubKey;
	Key privKey;
	public ELGAMAL_Cipher() throws Exception{
		Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
		this.cipher = Cipher.getInstance("ElGamal/None/NoPadding", "BC");//BC is the Provider, it stands for BouncyCastleProvider 
	    KeyPairGenerator generator = KeyPairGenerator.getInstance("ElGamal", "BC");
	    this.random = new SecureRandom();
	    generator.initialize(8*PRISM_Settings.ELGAMAL_KEY_SIZE, random);//key size = 256, ELGAMAL generator requires the key size to be specified in terms of # bits

	    KeyPair pair = generator.generateKeyPair();
	    this.pubKey = pair.getPublic();
	    this.privKey = pair.getPrivate();
	}
	
	Cipher getCipher(){
		return this.cipher;
	}
	SecureRandom getRandom(){
		return this.random;
	}
	Key getpubKey(){
		return this.pubKey;
	}
	Key getprivKey(){
		return this.privKey;
	}
	public static void main(String[] args) throws Exception {
		
		ELGAMAL_Cipher EGCipher=new ELGAMAL_Cipher();
		Cipher cipher = EGCipher.getCipher();
		
		byte[] zero = new byte[]{(byte)0x0}; //This is equal to 0
		byte[] one = PRISM_Settings.ONE; //This is equal to 1
		
	    String[] input = {"ab","ab"};//I use two identical strings to test the uniqueness of El gamal cipher's output
	    
	    for (String s:input){
		    cipher.init(Cipher.ENCRYPT_MODE, EGCipher.getpubKey(), EGCipher.getRandom());
		    byte[] cipherText = cipher.doFinal(s.getBytes());
		    System.out.println("cipher: " + new String(cipherText));
		    System.out.println("NumBytes:"+cipherText.length);
		    System.out.println("Bytes"+Arrays.toString(cipherText));
		    
		    cipher.init(Cipher.DECRYPT_MODE, EGCipher.getprivKey());
		    byte[] plainText = cipher.doFinal(cipherText);
		    System.out.println("plain : " + new String(plainText));
		    System.out.println("plaintext length: "+plainText.length+" ciphered text length: "+cipherText.length);
	    }
	    System.out.println("Bloc size :"+cipher.getBlockSize());
	    
	    cipher.init(Cipher.ENCRYPT_MODE, EGCipher.getpubKey(), EGCipher.getRandom());
	    byte [] zero_enc = cipher.doFinal(zero);
	    cipher.init(Cipher.ENCRYPT_MODE, EGCipher.getpubKey(), EGCipher.getRandom());
	    byte [] one_enc = cipher.doFinal(one);
	    BigInteger bi_zero_enc = new BigInteger(zero_enc);
	    BigInteger bi_one_enc = new BigInteger(one_enc);
	    
	    BigInteger bi_sum = bi_zero_enc.add(bi_one_enc);
	    BigInteger bi_product = bi_zero_enc.multiply(bi_one_enc); // this produce an output of size = product of input sizes
	    
	    System.out.println("E(0)= "+ Arrays.toString(bi_zero_enc.toByteArray()));
	    System.out.println("E(0) + E(1) = "+ Arrays.toString(bi_sum.toByteArray()));
	    System.out.println("E(0) * E(1) = "+ Arrays.toString(bi_product.toByteArray()));
	    
	    
	    cipher.init(Cipher.DECRYPT_MODE, EGCipher.getprivKey());
	    byte [] zero_dec = cipher.doFinal(zero);
	    cipher.init(Cipher.DECRYPT_MODE, EGCipher.getprivKey());
	    byte [] one_dec = cipher.doFinal(one);
	    System.out.println("zero_dec size : "+zero_dec.length);
	    System.out.println("zero_enc size : "+zero_enc.length);
	    
	    System.out.println("one_dec size : "+one_dec.length);
	    System.out.println("one_enc size : "+one_enc.length);
	    
	    byte[] sum_enc = bi_sum.toByteArray();
	    byte[] product_enc = bi_product.toByteArray();
	    System.out.println("sum size : "+sum_enc.length);
	    System.out.println("product size : "+product_enc.length);
	    
	    cipher.init(Cipher.DECRYPT_MODE, EGCipher.getprivKey());
	    byte[] result_sum = cipher.doFinal(sum_enc);
	    
	    // If I try to decrypt directly product_enc a DataLengthException:  attempt to process message too long for cipher
	    byte[] buf = new byte[PRISM_Settings.ELGAMAL_BLOCK_SIZE];
	    for(int pos =0;pos<product_enc.length;pos+=PRISM_Settings.ELGAMAL_BLOCK_SIZE){
	    	cipher.init(Cipher.DECRYPT_MODE, EGCipher.getprivKey());
	    	System.arraycopy(product_enc, pos, buf, 0, buf.length);
	    	System.out.println("buf size : "+buf.length);
	    	cipher.update(buf);
	    }
	    byte[] result_product = cipher.doFinal();
	    
	    //System.out.println("0 ="+  (new BigInteger(zero_dec)).intValue()); // Exception : Zero length BigInteger
	    System.out.println("0 + 1 = "+  (new BigInteger(result_sum)).intValue());
	    System.out.println("0 * 1 = "+  (new BigInteger(result_product)).intValue());
	    
	    
	    byte[] two = new byte[]{0x2};
	    byte[] num = new byte[]{0x10,0x10}; // == 4112
	    BigInteger bi;
	    BigInteger b1 = new BigInteger(one);
	    BigInteger b2 = new BigInteger(two);
	    BigInteger bi_num = new BigInteger(num);
	    
	    
	    bi = b1.add(b2);
	    System.out.println("b1 + b2 == "+bi.intValue());
	    bi = b1.multiply(b2);
	    System.out.println("b1 * b2 == "+bi.intValue());
	    bi = b1.add(bi_num);
	    System.out.println("b1 + num == "+bi.intValue());
	    System.out.println("b1 size: "+b1.toByteArray().length);
	    System.out.println("bi_num size: "+bi_num.toByteArray().length);
	    System.out.println("b1+num size: "+bi.toByteArray().length);
	  }
}
