import java.io.*;
import java.security.InvalidKeyException;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.Calendar;

import gnu.crypto.cipher.*;
import gnu.crypto.mac.*;
import gnu.crypto.pad.IPad;
import gnu.crypto.pad.PadFactory;

 
 public class RIJNDAEL_Cipher {
   
   private IBlockCipher cipher;
   
   /**
    * 
    * @param key_string : the secret key for this cipher
    * @param day_string : string representing the day
    * @throws Exception
    */
   public RIJNDAEL_Cipher(String key_string,String day_string) throws Exception{

           	   
	   String secret_key="";
	   	 
	   
		 if( key_string.length() >= PRISM_Settings.RJINDAEL_KEY_SIZE ){ // key size 32 bytes
				 secret_key=key_string;
				 
		 }else { 
			 System.err.println("ERROR key string should be at least :"+PRISM_Settings.RJINDAEL_KEY_SIZE+" bytes");
			 System.err.println("Aborting");
			 System.exit(1);
	 	 }
		 
	   	 byte[] key_bytes = computeKd(secret_key,day_string);
		// Creation of the cipher, specification of the algorithm and block size used
		 this.cipher = CipherFactory.getInstance("RIJNDAEL");
		 Map<String, Object> attributes = new HashMap<String, Object>();
	     attributes.put(IBlockCipher.KEY_MATERIAL, key_bytes);
	     attributes.put(IBlockCipher.CIPHER_BLOCK_SIZE, new Integer(PRISM_Settings.RIJNDAEL_BLOCK_SIZE));
	     this.cipher.init(attributes);
   }
   
   public RIJNDAEL_Cipher(String day_string) throws Exception{
	   this(PRISM_Settings.RIJNDAEL_DEFAULT_KEY_STRING,day_string);
	   System.err.println("Using the default key string: "+PRISM_Settings.RIJNDAEL_DEFAULT_KEY_STRING);
   }
   
   IBlockCipher getCipher(){
	   return this.cipher;
   }
   
   /**
    * Return an array of bytes made of the bytes in b properly padded to fit the block size
    * @param b input bytes
    * @param blocksize
    * @return padded bytes to fit the block size
    */
   public static byte[] padBytes(byte[] b, int blocksize){
	// Instanciation of the padding
	     // The word concatenated to the counter are transformed to fit the 32 bytes block size
	     IPad padding = PadFactory.getInstance("PKCS7");
	     padding.init(blocksize);
	     byte[] pad = padding.pad(b, 0, b.length);
	     byte[] b_padded = new byte[b.length + pad.length];
	     
	     System.arraycopy(b, 0, b_padded, 0, b.length);
	     System.arraycopy(pad, 0, b_padded, b.length, pad.length); //Copy pad.length bytes into b_padded starting from b.length
	     
	     return b_padded;
   }
   
   public byte[] unpadBytes(){
	    /*byte[] encrypted = new byte[b_padded.length];
	    byte[] plain_padded = new byte[b_padded.length];
	    */
	   return null;
   }
   
   /**
    * Computes the Key of the day Kd
    * @param key_str : a secret key string
    * @param  day_str : the date in the form dd/mm/yyyy
    * @return secret key of the day : Kd= HMAC(day) computed with key_str
    * @throws InvalidKeyException
    * @throws IllegalStateException
    */
   public static byte [] computeKd(String key_str,String day_str) throws InvalidKeyException, IllegalStateException{
	   byte[] key_bytes = key_str.getBytes();
	   	 Date day = null;
		 
		 DateFormat df = new SimpleDateFormat("dd-MM-yyyy");
		 try
		 {
			 day = df.parse(day_str);           
			 System.out.println("Date = " + df.format(day));
		 } catch (ParseException e)
		 {
			 e.printStackTrace();
			 System.out.println("Wrong date format please enter date : dd-MM-yyyy " );
			 System.exit(1);
		 }
		 
		 byte[] day_bytes = df.format(day).getBytes();
		 
		 // Instantiation of the HMAC, which output will by on 32 bytes = 256 bits as we are using SHA-256.
	     IMac mac = MacFactory.getInstance("HMAC-SHA-256");
	     HashMap<String, Object> attributes = new HashMap<String, Object>();
	     attributes.put(IMac.MAC_KEY_MATERIAL, key_bytes);
	     attributes.put(IMac.TRUNCATED_SIZE, new Integer(32));
	     mac.init(attributes);
	     mac.update(day_bytes, 0, day_bytes.length);
	     byte[] Kd = mac.digest();
	     
	     /*
	     System.out.println("Kd: " + Kd.toString());
	     System.out.println("Kd size: " + Kd.length);
	     */
	     return Kd;
   }
   
   public static void main(String[] args) throws Exception {
	   
	   
       byte[] dummy_buffer = new byte [PRISM_Settings.RJINDAEL_KEY_SIZE];


		 if(args.length < 2){
			 System.out.println("Usage: RIJNDAEL_Cipher <input_plain_padded_text_file> <ouput_ciphered_text_file> [<Secret key K>]");
			 System.exit(1);
		 }
		 
		 if(args.length >= 3 && args[2].length() < PRISM_Settings.RJINDAEL_KEY_SIZE){
			 System.out.println("The Secret key you entered is too short");
			 System.out.println("The length of the key must be at least the length of the computed HMAC (here, 32 bytes).");
			 System.out.println("[for testing purpose] The secret key is an optional parameter, if ommited a default key string will be used.");
			 System.exit(1);
		 }
		 
		 String key_str;
		 
		 if(args.length >= 3 && args[2].length() >= PRISM_Settings.RJINDAEL_KEY_SIZE ){ // key size 32 bytes
			 key_str=args[2];
			 
		 }else { // Use the default key string
			 System.out.println("Using the default key string: "+PRISM_Settings.RIJNDAEL_DEFAULT_KEY_STRING);
			 // Creation of Kd
			 key_str = PRISM_Settings.RIJNDAEL_DEFAULT_KEY_STRING;
			 /* For testing we use the fixed string:  "allyourbasearebelongtoushahahaha"
			 The length of the key must be at least the length of the computed hash (here, 32 bytes).
			 */
   		}
		 
		 String day_str = PRISM_Utils.getCurrentDate();
	     
	     /*
	     System.out.println("Kd: " + Kd.toString());
	     System.out.println("Kd size: " + Kd.length);
	     */
		 
	     //To use a simple predefined key
		 RIJNDAEL_Cipher RCipher = new RIJNDAEL_Cipher(key_str,day_str);
		 
		 IBlockCipher cipher = RCipher.getCipher();
	     int bs = cipher.currentBlockSize();

	     IPad padding = PadFactory.getInstance("PKCS7");
	     padding.init(bs);
	     
	     try {
	    	 File file_input = new File(args[0]);
	    	 BufferedReader reader = null;
	    	 reader = new BufferedReader(new FileReader(file_input));
	    	 
	    	 File file_output = new File(args[1]);
	    	 FileOutputStream out = new FileOutputStream(file_output);
			 
			 String line = null;
			 long j=0;
			 Hashtable<String, Integer> words_counters = new Hashtable<String, Integer>();
			 
			 // repeat until all lines of the plain_padded text are read
			 while ((line = reader.readLine()) != null){
				 
				 if (line.length()>0){
				 
					 String[] words = line.split(" ");
					 
					 for (int i = 0; i < words.length; i++){
						 
						 if (words[i].length() > PRISM_Settings.RIJNDAEL_BLOCK_SIZE - PRISM_Settings.COUNTER_LENGTH){
						     words[i] = words[i].substring(words[i].length()-PRISM_Settings.RIJNDAEL_BLOCK_SIZE + PRISM_Settings.COUNTER_LENGTH+1); 
   	 } 
						 //else
						 j++;
						 if ((j)%3==0)
 {						 
						  
							 int counter;
							 byte[] word_counter;  // This new word will be formed by the word input concatenated to its counter
							 word_counter = new byte[words[i].length() + PRISM_Settings.COUNTER_LENGTH]; // The counter is a number coded on 2 bits
							 
							 // The bytes in java are signed, the resulting integer is between 2^15 - 1 and 2^15 + 1

							 //System.out.println(words[i]);

							 byte[] word_byte = words[i].getBytes();
						     for (int k = 0; k < words[i].length(); k++)
						    	 word_counter[k] = word_byte[k];
						     
						     if (words_counters.containsKey(words[i])){
						    	 counter = (Integer) words_counters.get(words[i]) + 1;
						    	 words_counters.put(words[i], counter);
						     }
						     
						     else {
						    	 counter = 1;
						    	 words_counters.put(words[i], 1);
						     }
						     
						     // The counter is put at the end of the word, on 2 bytes
						     word_counter[words[i].length()+1] = (byte) (0xff & counter);
						     word_counter[words[i].length()] = (byte) (((0xff << 8) & counter) >> 8);
						     
						     // Instanciation of the padding
						     // The word concatenated to the counter are transformed to fit the 32 bytes block size
						     byte[] pad = padding.pad(word_counter, 0, word_counter.length);
						     byte[] word_padded = new byte[word_counter.length + pad.length];
						     byte[] encrypted = new byte[word_padded.length];
						     byte[] plain_padded = new byte[word_padded.length];
						     System.arraycopy(word_counter, 0, word_padded, 0, word_counter.length);
						     System.arraycopy(pad, 0, word_padded, word_counter.length, pad.length);
						     
						     // Encryption
						     cipher.encryptBlock(word_padded, 0, encrypted, 0);
						     
						     /**OUTPUT FILE**/
						     // We write the crypted word into the output file


							 out.write(encrypted);

						     // Decryption part, to verify the correctness of the program.
						     
						     /*
						      
						     cipher.decryptBlock(encrypted, 0, plain_padded, 0);
						     // We remove the padding from the decrypted word
						     int unpad = padding.unpad(plain_padded, 0, plain_padded.length);
						     byte[] plain = new byte[plain_padded.length - unpad];
						     System.arraycopy(plain_padded, 0, plain, 0, plain.length);
						     // I split the decrypted byte into the bytes corresponding to the word, and the ones corresponding to the counter.
						     byte[] word_decrypted = new byte[plain.length - PRISM_Settings.COUNTER_LENGTH];
						     byte[] counter_decrypted = new byte[2];
						     for (int k = 0; k < word_decrypted.length; k++)
						    	 word_decrypted[k] = plain[k];
						     for (int k = word_decrypted.length; k < plain.length; k++)
						    	 counter_decrypted[k - word_decrypted.length] = plain[k];
						     // Conversion of the counter's bytes into an int.
						     counter = counter_decrypted[0];
						     counter = (counter << 8) | (counter_decrypted[1] & 0xff);
						     String plainText = new String(word_decrypted); // Transform the byte[] into string 
						     System.out.println("Original text: " + plainText + " <--" + Arrays.toString(encrypted));
						     System.out.println("Counter: " + counter);
						     */
 } else out.write(dummy_buffer);
					    
						 }
					 }
				 }
			 System.out.println("Encryption completed");
			 reader.close();
			 out.close();
			 }
	     catch (FileNotFoundException e) {
	    	 e.printStackTrace();
	    	 System.exit(1);
	    	 } 
	     catch (IOException e) {
	    	 e.printStackTrace();
	    	 System.exit(1);
	    	 }
	     }
   }